import dataclasses
from typing import TYPE_CHECKING

import valohai_yaml

from papi import YAMLContext
from papi.base import PapiObject
from papi.contexts import APIContext

if TYPE_CHECKING:
    from papi.nodes import Node


@dataclasses.dataclass(frozen=True)
class Edge(PapiObject):
    source_node: str
    source_type: str
    source_key: str
    target_node: str
    target_type: str
    target_key: str
    configuration: dict = dataclasses.field(default_factory=dict)

    def to_api(self, api_context: APIContext):
        payload = dataclasses.asdict(self)
        if not payload.get("configuration"):
            del payload["configuration"]
        return payload

    def to_yaml(self, yaml_context: YAMLContext):
        return valohai_yaml.objs.Edge(
            source=".".join((self.source_node, self.source_type, self.source_key)),
            target=".".join((self.target_node, self.target_type, self.target_key)),
            configuration=self.configuration,
        )


@dataclasses.dataclass(frozen=True)
class EdgeDescriptor:
    node: "Node"
    type: str
    key: str

    def to(self, dest: "EdgeDescriptor"):
        # TODO: add validation
        assert self.node.papi is dest.node.papi
        self.node.papi.edges.append(
            Edge(
                source_node=self.node.name,
                source_type=self.type,
                source_key=self.key,
                target_node=dest.node.name,
                target_type=dest.type,
                target_key=dest.key,
            )
        )



from abc import ABCMeta
from typing import Any, Dict, Optional

import valohai_yaml
from valohai_yaml.objs import Step
from valohai_yaml.utils import listify

from .base import PapiObject
from .edges import EdgeDescriptor
from .utils import compact_dict


class Node(PapiObject, metaclass=ABCMeta):
    type: str
    name: str

    def __init__(self, *, name: str):
        self.name = name

    def output(self, key: str) -> EdgeDescriptor:
        return EdgeDescriptor(node=self, type="output", key=key)

    def input(self, key: str) -> EdgeDescriptor:
        return EdgeDescriptor(node=self, type="input", key=key)


class ExecutionNode(Node):
    type = "execution"
    parameters: Dict[str, Any]
    inputs: Dict[str, Any]

    def __init__(self, *, name: str, step: Step):
        super().__init__(name=name)
        self.step = step
        self.parameters = {}
        self.inputs = {}

    def parameter(self, name: str, value) -> "ExecutionNode":
        self._check_parameter_name(name)
        self.parameters[name] = value
        return self

    def to_api(self, api_context):
        step = self.step
        step_data = step.serialize()
        step_data.pop("name")
        step_data["parameters"] = {
            **step.get_parameter_defaults(include_flags=True),
            **self.parameters,
        }
        step_data["inputs"] = {
            i["name"]: self.inputs.get(i["name"], listify(i.get("default")))
            for i in step_data.get("inputs", [])
        }
        step_data.pop("mounts", None)
        return {
            "type": self.type,
            "name": self.name,
            "template": {
                "commit": api_context.commit_identifier,
                "step": step.name,
                **step_data,
            },
        }

    def to_yaml(self, yaml_context):
        return valohai_yaml.objs.ExecutionNode(
            name=self.name,
            step=self.step.name,
        )

    def input(self, key: str) -> EdgeDescriptor:
        if key not in self.step.inputs:
            raise ValueError(f"{key} not known in {self.step.name}")
        return super().input(key)

    def _check_parameter_name(self, name):
        if name not in self.step.parameters:
            raise ValueError(f"No such parameter {name}")


class TaskNode(ExecutionNode):
    type = "task"
    task_type: str = "grid_search"
    task_stop_expression: Optional[str] = None
    task_configuration: Optional[dict] = None

    def grid_search(self) -> "TaskNode":
        self.task_type = "grid_search"
        self.task_configuration = {}
        return self

    def random_search(self, executions: int) -> "TaskNode":
        self.task_type = "random_search"
        self.task_configuration = {
            "execution_count": executions,
        }
        return self

    def bayesian_tpe(self) -> "TaskNode":
        raise NotImplementedError("bayesian_tpe() is not implemented yet")

    def linear_parameter(
        self, name: str, *, min: float, max: float, step: float = 1
    ) -> "TaskNode":
        self._check_parameter_name(name)
        self.parameters[name] = {
            "style": "linear",
            "rules": {
                "start": float(min),
                "end": float(max),
                "step": float(step),
            },
        }
        return self

    def multiple_parameter(
        self,
        name: str,
        choices: list,
    ) -> "TaskNode":
        self._check_parameter_name(name)
        self.parameters[name] = {
            "style": "multiple",
            "rules": {
                "items": choices,
            },
        }
        return self

    def random_parameter(
        self,
        name: str,
        *,
        min: float,
        max: float,
        count: Optional[int] = 1,
        integerify: bool = False,
        seed: Optional[int] = None,
    ) -> "TaskNode":
        self._check_parameter_name(name)
        self.parameters[name] = {
            "style": "random",
            "rules": {
                "min": min,
                "max": max,
                "count": count,
                "integerify": bool(integerify),
                "seed": seed,
            },
        }
        return self

    def to_api(self, api_context):
        payload = super().to_api(api_context)
        payload["template"].update(
            compact_dict(
                dict(
                    type=self.task_type,
                    task_stop_expression=self.task_stop_expression,
                    configuration=self.task_configuration,
                )
            )
        )
        return payload

    def to_yaml(self, yaml_context):
        return valohai_yaml.objs.TaskNode(
            name=self.name,
            step=self.step.name,
        )



from collections import Counter
from typing import Dict, List, Optional, TypeVar

from valohai_yaml.objs import Config, Pipeline, Step

from papi.contexts import APIContext, YAMLContext
from papi.edges import Edge
from papi.nodes import ExecutionNode, Node, TaskNode

__all__ = ["Papi"]

SomeNode = TypeVar("SomeNode", bound=Node)


class Papi:
    nodes: Dict[str, Node]
    edges: List[Edge]
    counters: Counter

    def __init__(self, config: Config, *, name: str = "Pipeline"):
        self.name = str(name)
        self.config = config
        self.nodes = {}
        self.edges = []
        self.counters = Counter()

    def _count(self, key: str) -> str:
        self.counters[key] += 1
        key_l = key.lower().replace(" ", "_")
        return f"{key_l}_{self.counters[key]}"

    def _register_node(self, node: SomeNode) -> SomeNode:
        if node.name in self.nodes:
            raise ValueError(f"Duplicate node {node}")
        self.nodes[node.name] = node
        node.papi = self
        return node

    def execution(self, step: str, name: str = None):
        step_object = self._get_step(step)
        return self._register_node(
            ExecutionNode(
                step=step_object, name=(name or self._count(step_object.name))
            )
        )

    def _get_step(self, name) -> Step:
        step_object: Optional[Step] = self.config.get_step_by(name=name)
        if not step_object:
            raise ValueError(f"no such step {name} in config")
        return step_object

    def task(self, step: str, name: str = None) -> TaskNode:
        step_object = self._get_step(step)
        return self._register_node(
            TaskNode(step=step_object, name=(name or self._count(step_object.name)))
        )

    def to_api(self, api_context: Optional[APIContext] = None):
        if not api_context:
            api_context = APIContext()
        return {
            "nodes": [n.to_api(api_context=api_context) for n in self.nodes.values()],
            "edges": [e.to_api(api_context=api_context) for e in self.edges],
        }

    def to_yaml(self, yaml_context: YAMLContext = None):
        if not yaml_context:
            yaml_context = YAMLContext()
        return Pipeline(
            name=self.name,
            nodes=[n.to_yaml(yaml_context=yaml_context) for n in self.nodes.values()],
            edges=[e.to_yaml(yaml_context=yaml_context) for e in self.edges],
        )



import abc
from typing import TYPE_CHECKING

from .contexts import APIContext, YAMLContext

if TYPE_CHECKING:
    import papi  # noqa


class PapiObject(metaclass=abc.ABCMeta):
    papi: "papi.Papi"

    def to_api(self, api_context: APIContext):
        raise NotImplementedError(f"{self} must implement to_api")

    def to_yaml(self, yaml_context: YAMLContext):
        raise NotImplementedError(f"{self} must implement to_yaml")
